from functools import wraps
from typing import Callable, Awaitable, Union
import inspect

from .dot_dict import DotDict


def payload(func) -> Union[Callable[..., DotDict], Callable[..., Awaitable[DotDict]]]:
    """
    Decorator for payload functions
    """
    if inspect.iscoroutinefunction(func):
        async def wrapper(event: str, *args, **kwargs) -> DotDict:
            res = await func(event=event, *args, **kwargs)
            return DotDict(res)
    else:
        def wrapper(event: str, *args, **kwargs) -> DotDict:
            return DotDict(func(event=event, *args, **kwargs))

    wrapper = wraps(func)(wrapper)

    return wrapper



from typing import Any

from vk_dev import Condition


class Payload(Condition):
    """
    Check payload params in message

    *with_value: are any keys
    with this value or not

    **kwargs: is key from kwargs
    in payload.
    If key's value is None then
    return True if key doesn't
    exist.
    If key's value is Ellipsis (...) then
    return True if key exist
    with some value.
    If key has any different value
    then check values

    Without some values in __init__
    just check is a payload in message
    or not
    """
    def __init__(self, *with_value: Any, **kwargs: dict):
        self._with_value = with_value
        self._kwargs = kwargs

    def code(self, event, pl):
        if 'payload' in event.object.message:

            payload = eval(event.object.message.payload)

            for value in self._with_value:
                if value not in payload.values():
                    return False
                    
            for key, value in self._kwargs.items():
                if value is None:
                    if key in payload:
                        return False
                elif key not in payload:
                    if value is not None:
                        return False
                else:
                    if (value is not Ellipsis) and (payload[key] != value):
                        return False

            return True

        return False



from __future__ import annotations
import ssl
import os
from typing import (
    Optional,
    IO, Any
)

import aiohttp

from .dot_dict import DotDict


class DocInfo:
    def __init__(self, response: dict) -> None:
        self.type = response['type']
        self.info = DotDict(response[self.type])

    def __repr__(self) -> str:
        return f"doc{self.info.owner_id}_{self.info.id}"


class Document:
    """
    Quick send and work with documents
    """
    def __rrshift__(self, api) -> Document:
        self.api = api
        return self

    @classmethod
    def to_message(cls, **kwargs) -> Document:
        """
        Documents fos messages
        """
        self = cls()
        self._method = 'docs.getMessagesUploadServer'
        self._params = kwargs
        return self

    @classmethod
    def to_wall_and_message(cls, **kwargs) -> Document:
        """
        Documents for walls and messages
        """
        self = cls()
        self._method = 'docs.getWallUploadServer'
        self._params = kwargs
        return self

    async def load(
            self, *,
            file: Optional[IO[Any]] = None,
            title: Optional[str] = None,
            tags: Optional[str] = None,
            return_tags: Optional[bool] = None
        ) -> DocInfo:
        """
        Load doc to vk.
        Params will be passed to `docs.save` method.

        """
        if title is None:
            title = os.path.basename(file.name)

        url = await self.api.request(self._method, self._params)
        url = url['upload_url']

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url,
                data={
                    "file": file
                },
                ssl=ssl.SSLContext()
            ) as r:
                file_info = await r.json()

        if 'error' in file_info:
            raise ValueError(
                f"Something wrong with your file\n{file_info}"
            )

        file_info = await self.api.docs.save(
            file=file_info['file'],
            title=title,
            tags=tags,
            return_tags=return_tags
        )

        return DocInfo(file_info)



from __future__ import annotations
import json
from textwrap import dedent
from typing import (
    Optional,
    Union,
    Generator, AsyncGenerator
)


class Button:
    """
    Keyboard button
    """
    def __init__(self, **kwargs) -> None:
        self.info = {'action': {**kwargs}}

    def positive(self) -> Button:
        """
        Green button
        """
        self.info['color'] = 'positive'
        return self

    def negative(self) -> Button:
        """
        Red button
        """
        self.info['color'] = 'negative'
        return self

    def secondary(self) -> Button:
        """
        White button
        """
        self.info['color'] = 'secondary'
        return self

    def primary(self) -> Button:
        """
        Blue button
        """
        self.info['color'] = 'primary'
        return self

    @classmethod
    def _button_init(cls, **kwargs) -> Button:
        self = super().__new__(cls)
        if 'payload' in kwargs:
            kwargs['payload'] = str(kwargs['payload']).replace('\'', '"')
        self.__init__(**kwargs)

        return self

    @classmethod
    def line(cls) -> Button:
        """
        Add Buttons line
        """
        self = cls._button_init()
        self.info = None

        return self

    @classmethod
    def text(cls, **kwargs) -> Button:
        return cls._button_init(type='text', **kwargs)

    @classmethod
    def open_link(cls, **kwargs) -> Button:
        return cls._button_init(type='open_link', **kwargs)

    @classmethod
    def location(cls, **kwargs) -> Button:
        return cls._button_init(type='location', **kwargs)

    @classmethod
    def vkpay(cls, **kwargs) -> Button:
        return cls._button_init(type='vkpay', **kwargs)

    @classmethod
    def open_app(cls, **kwargs) -> Button:
        return cls._button_init(type='open_app', **kwargs)


class Keyboard:
    """
    messages.send: keyboard
    Create VK Keyboard by dict or buttons list
    """
    def __init__(self, kb: Optional[dict] = None, **kwargs) -> None:
        if kb is None:
            self.kb = {
                **kwargs,
                'buttons': [[]]
            } if kwargs else []
        else:
            self.kb = kb

        # Check last for template or not
        self._kwargs = kwargs

    def __add__(self, button: Button) -> None:
        """
        Add button to line
        """

        self.kb['buttons'][-1].append(button())

    def __repr__(self) -> str:
        """
        Create for sending
        """
        kb = json.dumps(self.kb, ensure_ascii=False)
        kb = kb.encode('utf-8').decode('utf-8')
        return str(kb)

    def create(self, *buttons: Button) -> Keyboard:
        """
        Create keyboard by Buttons object
        """
        for button in buttons:
            if not isinstance(button, Button):
                error_msg = "Keyboard's buttons must be 'Button' instance, not"
                raise TypeError(
                    dedent(f"""
                        {error_msg} '{type(button).__name__}'
                    """)
                )
            if button.info is None:
                self.kb['buttons'].append([])
            else:
                if self._kwargs:
                    self.kb['buttons'][-1].append(button.info)
                else:
                    self.kb.append(button.info)

        return self


class Element:
    """
    Element for carousel
    """
    def __init__(self, **kwargs) -> None:
        self.info = kwargs
        self.info['action'] = {}

    def open_link(self, link: str) -> "Element":
        self.info['action']['type'] = 'open_link'
        self.info['action']['link'] = link
        return self

    def open_photo(self) -> "Element":
        self.info['action']['type'] = 'open_photo'
        return self


class Template:
    """
    messages.send: template
    For example, you can
    create carousel with this
    """
    def __init__(self, tp: Optional[dict] = None) -> None:
        """
        Init by dict described by vk scheme.
        Also you can use generators for this,
        just decorate your generator
        with Template's object
        """
        if tp is None:
            self.tp = {
                'type': 'carousel',
                'elements': []
            }
        else:
            self.tp = tp

    def __call__(
            self,
            gen: Union[
                Generator[Element, None, None],
                AsyncGenerator[Element, None]
            ]
        ) -> None:
        """
        Create carousel with
        decorating generatore
        """
        for elem in gen():
            elem.info['buttons'] = elem.info['buttons'].kb
            self.tp['elements'].append(elem.info)

    def __repr__(self) -> str:
        tp = json.dumps(self.tp, ensure_ascii=False)
        tp = tp.encode('utf-8').decode('utf-8')
        return str(tp)



class DotDict(dict):
    """
    Wrapper for dictioanries for calling elements by dot,
    also work like ordinary dict.
    Work correctly if all elements are primitives
    """

    def __init__(self, d):
        super().__init__(d)

    def __getattr__(self, value):
        """
        Congested for dot-syntax
        """
        return self._get_value(self[value])

    def _get_value(self, value):
        """
        Return dict that wrapped in DotDict
        """

        if isinstance(value, dict):
            return self.__class__(value)

        elif isinstance(value, list):
            return [self._get_value(i) for i in value]

        return value



from vk_dev import Condition


class Prefix(Condition):
    """
    Check how message start
    """

    def __init__(self, *prefixes: str):
        self._prefixes = prefixes

    def code(self, event, pl):
        if event.object.message.text.startswith(self._prefixes):
            return True
        return False



import uuid
from typing import Callable, Awaitable, Union
from abc import ABC, abstractmethod


class Condition(ABC):
    """
    Inheriting gives you new decorators
    """
    def __call__(self, func) -> Union[Callable[[str, str], None], Callable[[str, str], Awaitable[None]]]:
        self.uuid = uuid.uuid4().hex
        self.func = func
        func.conditions.append(self)
        if not hasattr(self, 'conf_id'):
            self.conf_id = self.uuid

        return func

    @abstractmethod
    def code(self, event, pl) -> bool:
        """
        Code that should return True/False.
        Calling with LP event and payload.
        ```python
        def code(self, event, pl):
            if event.object.message.text.startswith(self._prefixes):
                return True
            return False
        ```
        """



from typing import Literal

from vk_dev import Condition


class By(Condition):
    """
    Check, who wrote a message: group or user
    """
    def __init__(self, writer: Literal['user', 'group']):
        self._writer_type = writer_type

    def code(self, event):
        if event.object.message.from_id > 0:
            return 'user' == self._writer_type
        return 'group' == self._writer_type



from .condition import Condition
from .tools import peer, random_id
from .exception import VkErr
from .payload import payload
from .dot_dict import DotDict
from .upload import (
    Document
)
from .api import (
    Api, LongPoll
)
from .vkui import (
    Keyboard, Button,
    Template, Element,
)

from vk_dev import cond



from typing import Literal

from vk_dev import Condition, peer


class Path(Condition):
    """
    Message's path
    """

    def __init__(self, path: Literal['chat', 'direct']):
        self._path = path

    def code(self, event, pl):
        if event.object.message.peer_id > peer:
            return 'direct' == self._path
        return 'chat' == self._path



from textwrap import dedent


class VkErr(Exception):
    """
    Errors from response from vk
    """
    lp_faileds = {
        1: "The event history went out of date or was partially lost",
        2: "The key’s active period expired",
        3: "Information was lost",
        4: "Invalid version number was passed"
    }
    lp_handler_text = dedent("""
        . \033[33mYou can use auto LP faileds handler in LongPoll class\033[0m
    """)

    def __init__(self, text) -> None:
        self.text = text

    @staticmethod
    def text_init(text) -> str:
        if isinstance(text, dict):
            if "failed" in text:
                return VkErr.lp_faileds[text["failed"]] + VkErr.lp_handler_text

            elif 'error' in text:
                error_code = text['error']['error_code']
                error_msg = text['error']['error_msg']
                head_msg = dedent(f"""
                    \n\033[31m[{error_code}] {error_msg}\033[0m\n
                """)

                content = "Request params:"
                for pair in text["error"]["request_params"]:
                    key = f"\n\033[33m{pair['key']}\033[0m"
                    value = f"\033[36m{pair['value']}\033[0m"
                    content += f"{key} = {value}"

                return head_msg + content

            return text

        return text



from __future__ import annotations
import time
import asyncio
import ssl
import inspect
from datetime import datetime as dt
from typing import (
    Union, Any,
    NoReturn
)

import aiohttp

from .exception import VkErr
from .dot_dict import DotDict


class Api:
    """
    Make async API requests and
    API-helpers handler.

    Contain main info like
    access_token,
    version and etc.
    """
    def __init__(
            self,
            token: str,
            v: Union[str, float],
            group_id: int = None
        ) -> None:
        self.url = 'https://api.vk.com/method/'
        self.token = token
        self.v = str(v)
        self.group_id = abs(group_id) if group_id is not None else group_id
        self.ssl = ssl.SSLContext()
        self._last_request_time = time.time()
        self._freeze_time = 0.05
        self._method = None

    def __getattr__(self, value: str) -> Api:
        self.method = value
        return self

    async def __call__(self, **kwargs) -> DotDict:
        """
        Make every API requests
        """
        return await self.request(self.method, data=kwargs)

    async def request(self, method: str, data: dict) -> DotDict:
        """
        Make every API requests
        """
        api_data = {
            'access_token': self.token,
            'v': self.v,
            **data
        }

        await self._request_wait()

        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.url + method, data=api_data, ssl=self.ssl
            ) as r:
                resp = await r.json()

        return self._error_check(resp)

    async def _request_wait(self) -> None:
        """
        Pause between requests
        """
        now = time.time()
        diff = now - self._last_request_time

        if diff < self._freeze_time:
            await asyncio.sleep(self._freeze_time - diff)
            self._last_request_time = time.time()

    def _error_check(self, resp: dict) -> Union[DotDict, Any]:
        """
        Check errors from response
        """
        if 'error' in resp:
            raise VkErr(VkErr.text_init(resp))
        else:
            if isinstance(resp['response'], dict):
                return DotDict(resp['response'])
            return resp['response']

    @property
    def method(self) -> None:
        """
        Method name
        """
    @method.getter
    def method(self) -> str:
        res = self._method
        self._method = None
        return res

    @method.setter
    def method(self, value: str) -> None:
        if self._method is None:
            self._method = value
        else:
            self._method += '.' + value


class LongPoll:
    """
    LongPoll scheme
    """
    group_get = {
        # group_id
    }
    group_init = {
        'wait': 25
    }

    def __init__(self, faileds=[1, 2, 3], default=True, **kwargs) -> None:
        self.faileds = faileds
        self.start_settings = kwargs
        self.reaction_handlers = []

    def __rrshift__(self, api):
        """
        Use to add your API configuration to LP
        """
        self.api = api
        return self

    def __getattr__(self, event_name: str) -> _ReactionHandler:
        """
        Get handling event
        """
        hand = _ReactionHandler(event_name)
        self.reaction_handlers.append(hand)

        return hand

    async def _lp_start(self, default=True, **kwargs) -> NoReturn:

        # Reactions tree
        self._reactions_init()

        LongPoll.group_get['group_id'] = self.api.group_id

        self.start_settings = {
                **(
                    (
                        LongPoll.group_get
                    ) if default else {}
                ),
                **self.start_settings
            }

        # Yours settings
        self.lp_settings = {
            **LongPoll.group_init,
            **kwargs
        } if default else kwargs

        # Intermediate lp params like server, ts and key
        self.lp_info = await self.api.request(
                method='groups.getLongPollServer',
                data=self.start_settings
            )
        self.start_time = dt.now()
        self.format_start = self.start_time.strftime("[%Y-%m-%d %H:%M:%S]")
        listening_string = "\033[0m\033[32mListening VK LongPoll...\033[0m"
        print(f"\033[2m{self.format_start} {listening_string}")

        # Stats
        self.events_get = 0
        self.events_handled = 0

        while True:
            # Lp events
            lp_get = {
                'key': self.lp_info['key'],
                'ts': self.lp_info['ts']
            }

            data = {**lp_get, **self.lp_settings, 'act': 'a_check'}

            async with aiohttp.ClientSession() as session:
                async with session.post(
                    self.lp_info['server'],
                    data=data,
                    ssl=self.api.ssl
                ) as response:
                    self.lp = await response.json()

            error = await self._failed_handler()
            if error is True:
                continue

            for update in self.lp['updates']:
                event = DotDict(update)
                self.events_get += 1
                if event.type in self.reactions:
                    self.events_handled += 1
                    asyncio.create_task(self._reactions_get(event))

    def __call__(self, default=True, **kwargs) -> NoReturn:
        """
        Init LongPoll listening
        """
        try:
            # Check `api >>` was
            if not hasattr(self, 'api'):
                raise ValueError("API config not founded. Make sure you use `api >> ` before LongPoll assignment")

            # Check it's a group 
            if self.api.group_id is None:
                raise ValueError("Can't use user's token for LongPoll now :(")

            # Infinity corutine
            asyncio.run(self._lp_start(default, **kwargs))

        except KeyboardInterrupt:
            end_time = dt.now()
            dif = end_time - self.start_time
            format_end = end_time.strftime("[%Y-%m-%d %H:%M:%S]")

            listening_stop = "\033[0m\033[33mListening has been stoped\033[0m"
            print(f"\n\033[2m{format_end} {listening_stop}")
            print("Handled \033[36m%s\033[0m (\033[35m%.2f ps\033[0m)" % (
                self.events_handled,
                self.events_handled / dif.seconds
            ))
            print("Total \033[36m%s\033[0m (\033[35m%.2f ps\033[0m)" % (
                self.events_get,
                self.events_get / dif.seconds
            ))
            print(f"Taken \033[36m{dif}\033[0m")

            exit()

    async def _reactions_call(self, event, reaction) -> None:
        """
        Call every reaction
        """
        payload = DotDict({})
        for pl_gen in reaction.pl_gen:
            if inspect.iscoroutinefunction(pl_gen):
                res = await pl_gen(event)
                payload = DotDict({**payload, **res})
            else:
                payload = DotDict({**payload, **pl_gen(event)})

        for cond in reaction.conditions:
            if inspect.iscoroutinefunction(cond.code):
                res = await cond.code(event, payload)
                if not res:
                    break
            else:
                if not cond.code(event, payload):
                    break
        else:
            if inspect.iscoroutinefunction(reaction):
                asyncio.create_task(reaction(event, payload))
            else:
                reaction(event, payload)

    async def _reactions_get(self, event) -> None:
        """
        Return list of needed funcs with payload
        """
        for reaction in self.reactions[event.type]:
            asyncio.create_task(self._reactions_call(event, reaction))

    def _reactions_init(self) -> None:
        """
        Init reactions tree
        """
        reactions = {}
        for handler in self.reaction_handlers:
            if handler.event_name not in reactions:
                reactions[handler.event_name] = [handler.reaction]
            else:
                reactions[handler.event_name].append(handler.reaction)

        self.reactions = reactions

    async def _failed_handler(self) -> Union[bool, None]:
        """
        Catch lp faileds
        """
        if 'failed' in self.lp:
            if self.lp['failed'] in self.faileds:
                await self._failed_resolving()
                return True
            else:
                raise VkErr(VkErr.text_init(self.lp))
        else:
            self.lp_info['ts'] = self.lp['ts']

    async def _failed_resolving(self) -> None:
        """
        Resolve faileds problems
        """
        if self.lp['failed'] == 1:
            self.lp_info['ts'] = self.lp['ts']
        elif self.lp['failed'] in (2, 3):
            self.lp_info = await self.api.request(
                    self._method_name(),
                    self.start_settings
                )


class _ReactionHandler:
    """
    Reactions Handler
    """
    def __init__(self, event_name: str) -> None:
        self.event_name = event_name

    def __call__(self, *pl_gen):
        """
        Take payload generator
        """
        self.pl_gen = pl_gen

        self.__class__.__call__, self.__class__._reaction_decor =\
            self.__class__._reaction_decor, self.__class__.__call__

        return self

    def _reaction_decor(self, func: Any) -> Any:
        """
        Called when it is decorating
        """
        self.__class__.__call__, self.__class__._reaction_decor =\
            self.__class__._reaction_decor, self.__class__.__call__
        self.reaction = func

        func.event_name = self.event_name
        func.conditions = []
        func.pl_gen = self.pl_gen

        return func
